"""OTEAPI Filter strategy config AiiDA Data Node class."""
# pylint: disable=invalid-name
from typing import TYPE_CHECKING

from oteapi.models import FilterConfig

from execflow.wrapper.data.genericconfig import GenericConfigData

if TYPE_CHECKING:  # pragma: no cover
    from typing import Any, Optional, Union


class FilterConfigData(GenericConfigData):
    """Filter Strategy Data Configuration.

    Args:
        filterType (str): Type of registered filter strategy. E.g., `filter/sql`.
        query (Optional[str]): Define a query operation.
        condition (Optional[str]): Logical statement indicating when a filter should be
            applied.
        limit (Optional[int]): Number of items remaining after a filter expression.

    """

    def __init__(
        self,
        filterType: str,
        query: "Optional[str]" = None,
        condition: "Optional[str]" = None,
        limit: "Optional[int]" = None,
        **kwargs: "Any",
    ) -> None:
        if query is None:
            query = FilterConfig.schema()["properties"]["query"]["default"]

        if condition is None:
            condition = FilterConfig.schema()["properties"]["condition"]["default"]

        if limit is None:
            limit = FilterConfig.schema()["properties"]["limit"]["default"]

        super().__init__(**kwargs)

        attr_dict = {
            "filterType": filterType,
            "query": query,
            "condition": condition,
            "limit": limit,
        }

        self.base.attributes.set_many(attr_dict)

    @property
    def filterType(self) -> str:
        """Type of registered filter strategy. E.g., `filter/sql`."""
        return self.base.attributes.get("filterType")

    @filterType.setter
    def filterType(self, value: str) -> None:
        self.set_attribute("filterType", value)

    @property
    def query(self) -> "Union[str, None]":
        """Define a query operation."""
        return self.base.attributes.get("query")

    @query.setter
    def query(self, value: "Union[str, None]") -> None:
        self.set_attribute("query", value)

    @property
    def condition(self) -> "Union[str, None]":
        """Logical statement indicating when a filter should be applied."""
        return self.base.attributes.get("condition")

    @condition.setter
    def condition(self, value: "Union[str, None]") -> None:
        self.set_attribute("condition", value)

    @property
    def limit(self) -> "Union[int, None]":
        """Number of items remaining after a filter expression."""
        return self.base.attributes.get("limit")

    @limit.setter
    def limit(self, value: "Union[int, None]") -> None:
        self.set_attribute("limit", value)



"""Test execflow.wrapper.workflows.pipeline"""
from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Dict, Tuple, Union


def get_input_variants() -> "Dict[str, Tuple[str, Union[Path, bytes, dict]]]":
    """Input for 'test_run_pipeline'.

    Must pass AiiDA input variants as tuples of entry point entries and initialization
    object to the AiiDA Data Node type.
    This is because the AiiDA Profile (DB/storage) is not created/initialized until
    actually entering a test. And this function is *not* a test.
    """
    from pathlib import Path

    import yaml

    samples = (Path(__file__).resolve().parent.parent.parent / "samples").resolve()

    return {
        "SinglefileData": ("core.singlefile", samples / "pipe.yml"),
        "Str path": ("core.str", str(samples / "pipe.yml")),
        "Str content": (
            "core.str",
            (samples / "pipe.yml").read_text(encoding="utf8"),
        ),
        "OTEPipelineData": (
            "execflow.otepipeline",
            (samples / "pipe.yml").read_bytes(),
        ),
        "orm.Dict": (
            "core.dict",
            yaml.safe_load((samples / "pipe.yml").read_bytes()),
        ),
    }


@pytest.mark.parametrize(
    ["entry_point", "node_input"],
    get_input_variants().values(),
    ids=get_input_variants().keys(),
)
def test_run_pipeline(entry_point: str, node_input: "Union[Path, str, dict]") -> None:
    """Run a simple pipeline.

    Parameters:
        entry_point: An AiiDA entry point for Data Nodes.
        node_input: Input for initializing the given Data Node.

    """
    from aiida.engine import run
    from aiida.plugins import DataFactory

    from execflow.wrapper.workflows.pipeline import OTEPipeline

    input_variant = DataFactory(entry_point)(node_input)
    run(OTEPipeline, pipeline=input_variant)


def test_result_pipeline(samples: "Path") -> None:
    """Run a simple pipeline and check result.

    Parameters:
        samples: Path to test directory with sample files.

    """
    import requests
    import yaml
    from aiida.engine import run
    from aiida.plugins import DataFactory

    from execflow.wrapper.workflows.pipeline import OTEPipeline

    entry_point, node_input = get_input_variants()["OTEPipelineData"]
    input_variant = DataFactory(entry_point)(node_input)
    result = run(OTEPipeline, pipeline=input_variant)

    assert len(result) == 1
    assert "session" in result
    assert isinstance(result["session"], DataFactory("core.dict"))

    declarative_pipeline_file = yaml.safe_load((samples / "pipe.yml").read_bytes())
    json_file = requests.get(
        declarative_pipeline_file["strategies"][0]["downloadUrl"],
        timeout=5,
    ).json()

    # The JSON parser strategy will add a "content" entry in the session.
    # This will contain the parsed content.
    assert "content" in result["session"]
    assert result["session"]["content"] == json_file

    # The mapping strategy will add "prefixes" and "triples" entries in the session.
    assert "prefixes" in result["session"]
    assert "triples" in result["session"]
    assert (
        result["session"]["prefixes"]
        == declarative_pipeline_file["strategies"][1]["prefixes"]
    )
    assert (
        result["session"]["triples"]
        == declarative_pipeline_file["strategies"][1]["triples"]
    )



# pylint: disable=missing-docstring,similarities,fixme,invalid-name,redefined-builtin
import execflow

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "ExecFlow SDK"
copyright = "© SINTEF 2022"
author = "SINTEF"
version = ".".join(execflow.__version__.split(".")[:2])  # Short X.Y version
release = execflow.__version__
# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration


# General configuration
extensions = [
    "myst_nb",  # markdown source support & support for Jupyter notebooks
    "autoapi.extension",
    "sphinx.ext.graphviz",
    "sphinx.ext.intersphinx",  # Connect to external Sphinx documentation
    "sphinx.ext.napoleon",  # API ref Google and NumPy style
    "sphinx.ext.viewcode",
    "sphinxcontrib.plantuml",
    "sphinx_copybutton",  # Copy button for codeblocks
    "sphinx_design",
]

autoapi_dirs = ["../execflow"]
autoapi_type = "python"
autoapi_file_patterns = ["*.py", "*.pyi"]
autoapi_template_dir = "_templates/autoapi"
autoapi_add_toctree_entry = True  # Toggle the most top-level index file
autoapi_options = [
    "members",
    "undoc-members",
    "private-members",
    "show-inheritance",
    "show-module-summary",
    "special-members",
    "imported-members",
    "show-inheritance-diagram",
    # "inherited-members",
]
autoapi_keep_files = False  # Should be False in production

autodoc_typehints = "description"
autodoc_typehints_format = "short"
autodoc_inherit_docstrings = True

root_doc = "index"

myst_heading_anchors = 5
myst_enable_extensions = ["colon_fence"]

plantuml = "java -jar lib/plantuml.jar"
plantuml_output_format = "svg_img"

exclude_patterns = [
    "_build",
    "Thumbs.db",
    ".DS_Store",
    "**.ipynb_checkpoints",
    "attic/**/*",
    "_templates",
]

# HTML output
html_theme = "sphinx_book_theme"
html_logo = "_static/logo.jpeg"
html_favicon = "_static/favicon.ico"
html_theme_options = {
    "path_to_docs": "docs",
    "repository_url": "https://github.com/H2020-OpenModel/WP3-WrapperSDK",
    "repository_branch": "master",
    "use_issues_button": True,
    "use_fullscreen_button": True,
    "use_repository_button": True,
    "logo_only": True,
    "show_navbar_depth": 1,
    # "announcement": "This is an announcement!",
}
html_static_path = ["_static"]
# html_css_files = ["custom.css"]

nbsphinx_allow_errors = True

suppress_warnings = ["myst.mathjax"]

napoleon_use_admonition_for_examples = True
napoleon_use_admonition_for_notes = True
napoleon_use_admonition_for_references = True
napoleon_preprocess_types = True
napoleon_attr_annotations = True

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "aiida": ("https://aiida.readthedocs.io/projects/aiida-core/en/latest", None),
    "oteapi": ("https://emmc-asbl.github.io/oteapi-core/latest/", None),
}



"""Pytest-specific file for fixtures and configuration.

Since this `conftest` file is at the root level of the `tests` folder, all fixtures
here are available for all tests.

Use AiiDA's pytest fixtures.
Look inside aiida.manage.tests.pytest_fixtures to see which fixtures are provided:
https://aiida.readthedocs.io/projects/aiida-core/en/latest/reference/apidoc/aiida.manage.tests.html#module-aiida.manage.tests.pytest_fixtures
"""
from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from pathlib import Path

pytest_plugins = ["aiida.manage.tests.pytest_fixtures"]


@pytest.fixture(scope="function", autouse=True)
@pytest.mark.usefixtures("aiida_profile_clean")
def aiida_profile_clean_auto():
    """Automatically clear the AiiDA profile's DB and storage after each test."""


@pytest.fixture(scope="session")
def samples() -> "Path":
    """Return path to 'samples' folder."""
    from pathlib import Path

    path = (Path(__file__).resolve().parent / "samples").resolve()
    if not path.exists():
        raise FileNotFoundError(f"Could not locate the 'samples' folder at: {path}")
    return path



"""AiiDA Calculations for the OTE Transformation Strategy."""
from time import sleep, time
from typing import TYPE_CHECKING

from aiida.engine import calcfunction
from aiida.orm import Dict
from oteapi.models import TransformationConfig
from oteapi.plugins import create_strategy, load_strategies

if TYPE_CHECKING:  # pragma: no cover
    from oteapi.interfaces import ITransformationStrategy
    from oteapi.models import TransformationStatus


@calcfunction
def init_transformation(config: Dict, session: Dict) -> Dict:
    """Initialize an OTE Transformation strategy."""
    load_strategies()

    transformation_config = TransformationConfig(**config.get_dict())
    strategy: "ITransformationStrategy" = create_strategy(
        "transformation", transformation_config
    )
    return Dict(strategy.initialize(session.get_dict()))


@calcfunction
def get_transformation(config: Dict, session: Dict) -> Dict:
    """Get an OTE Transformation strategy.

    Important:
        Currently, the status values are valid only for Celery.

        This is because only a single transformation strategy exists (for Celery) and
        the configuration and status models have been based on this strategy.

        A status enumeration should be set as the type for
        `TransformationStatus.status` in order to more agnostically determine the state
        from any transformation strategy.

        However, this is to be implemented in OTEAPI Core.

    """
    load_strategies()

    transformation_config = TransformationConfig(**config.get_dict())
    strategy: "ITransformationStrategy" = create_strategy(
        "transformation", transformation_config
    )

    wall_time = 2 * 60  # 2 min.

    start_time = time()
    status: "TransformationStatus" = strategy.run(session.get_dict())
    while (
        status.status
        not in (
            "READY_STATES",
            "EXCEPTION_STATES",
            "PROPAGATE_STATES",
            "SUCCESS",
            "FAILURE",
            "REVOKED",
            "RETRY",
        )
        or not status.finishTime
    ) and (time() - start_time) < wall_time:
        sleep(0.5)
        status = strategy.status(status.id)

    return Dict(strategy.get(session.get_dict()))



"""Test execflow.wrapper.data.declarative_pipeline"""
# pylint: disable=too-many-locals,invalid-name
from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Any, Dict, Tuple, Type, Union

    from execflow.wrapper.data.declarative_pipeline import (
        OTEPipelineData as OTEPipelineDataNode,
    )


@pytest.mark.parametrize(
    "value_type",
    (
        dict,
        "execflow.otepipeline",
        "core.dict",
        "DeclarativePipeline",
        str,
        bytes,
        "core.str",
    ),
)
def test_initialization_strategies_value(
    value_type: "Union[Type[Union[dict, str, bytes]], str]", samples: "Path"
) -> None:
    """Ensure the different intended initialization strategies work.

    This test is for the `value` pathway, i.e., using the `value` parameter.

    Parameters:
        value_type: One of the valid types for the `value` parameter.
        samples: Path to test directory with sample files.

    """
    from copy import deepcopy

    import yaml
    from aiida.plugins import DataFactory

    from execflow.utils import DeclarativePipeline

    OTEPipelineData: "Type[OTEPipelineDataNode]" = DataFactory("execflow.otepipeline")
    declarative_pipeline_file: "Dict[str, Any]" = yaml.safe_load(
        (samples / "pipe.yml").read_bytes()
    )

    if isinstance(value_type, type):
        if value_type == dict:
            pipeline_input = deepcopy(declarative_pipeline_file)
            assert isinstance(pipeline_input, dict)
            node = OTEPipelineData(value=pipeline_input)
        elif value_type == str:
            node = OTEPipelineData(
                value=(samples / "pipe.yml").read_text(encoding="utf8")
            )
        else:
            assert value_type == bytes, f"Unknown value_type type: {value_type}"
            node = OTEPipelineData(value=(samples / "pipe.yml").read_bytes())
    else:
        # value_type is a string
        if value_type == "DeclarativePipeline":
            pydantic_input = deepcopy(declarative_pipeline_file)
            pydantic_model = DeclarativePipeline(**pydantic_input)
            node = OTEPipelineData(value=pydantic_model)
        elif value_type == "execflow.otepipeline":
            pipeline_data_node = OTEPipelineData()
            pipeline_data_node.version = declarative_pipeline_file["version"]
            pipeline_data_node.strategies = declarative_pipeline_file["strategies"]
            pipeline_data_node.pipelines = declarative_pipeline_file["pipelines"]
            node = OTEPipelineData(value=pipeline_data_node)
        elif value_type == "core.dict":
            dict_node_input = deepcopy(declarative_pipeline_file)
            aiida_dict_node = DataFactory(value_type)(dict_node_input)
            node = OTEPipelineData(value=aiida_dict_node)
        else:
            assert value_type == "core.str", f"Unknown value_type given: {value_type}"
            aiida_str_node = DataFactory(value_type)(
                (samples / "pipe.yml").read_text(encoding="utf8")
            )
            node = OTEPipelineData(value=aiida_str_node)

    assert node.validate() is None
    assert node.version == int(declarative_pipeline_file["version"])
    assert node.pipelines == dict(declarative_pipeline_file["pipelines"])

    if value_type == "DeclarativePipeline":
        # A lot of default values will be added to the strategies from the pydantic
        # model(s).
        assert node.strategies != list(declarative_pipeline_file["strategies"])

        # Initialize pydantic model and convert strategies to a list of dictionaries.
        pydantic_input = deepcopy(declarative_pipeline_file)
        pydantic_model = DeclarativePipeline(**pydantic_input)
        assert node.strategies == [_.dict() for _ in pydantic_model.strategies]
    else:
        assert node.strategies == list(declarative_pipeline_file["strategies"])


@pytest.mark.parametrize(
    "input_type",
    (
        "Path",
        str,
        "core.str",
        "core.singlefile",
    ),
)
def test_initialization_strategies_file(
    input_type: "Union[Type[str], str]", samples: "Path"
) -> None:
    """Ensure the different intended initialization strategies work.

    This test is for the file pathway, i.e., using the `filepath` or `single_file`
    parameter.

    Parameters:
        input_type: One of the valid types for the `filepath` or `single_file`
            parameter.
        samples: Path to test directory with sample files.

    """
    import yaml
    from aiida.plugins import DataFactory

    OTEPipelineData: "Type[OTEPipelineDataNode]" = DataFactory("execflow.otepipeline")
    declarative_pipeline_file: "Dict[str, Any]" = yaml.safe_load(
        (samples / "pipe.yml").read_bytes()
    )

    if input_type == str:
        # filepath
        node = OTEPipelineData(filepath=str(samples / "pipe.yml"))
    else:
        assert isinstance(input_type, str)

        # filepath
        if input_type == "Path":
            node = OTEPipelineData(filepath=samples / "pipe.yml")
        elif input_type == "core.str":
            aiida_str_node = DataFactory(input_type)(str(samples / "pipe.yml"))
            node = OTEPipelineData(filepath=aiida_str_node)

        # single_file
        else:
            assert (
                input_type == "core.singlefile"
            ), f"Unknown input_type given: {input_type}"

            aiida_singlefile_node = DataFactory(input_type)(samples / "pipe.yml")
            node = OTEPipelineData(single_file=aiida_singlefile_node)

    assert node.validate() is None
    assert node.version == int(declarative_pipeline_file["version"])
    assert node.pipelines == dict(declarative_pipeline_file["pipelines"])
    assert node.strategies == list(declarative_pipeline_file["strategies"])


def test_initialization_strategies_explicit(samples: "Path") -> None:
    """Ensure the different intended initialization strategies work.

    This test is for explicitly filling out the top keywords.
    I.e., `version`, `strategies`, and `pipelines`.

    Parameters:
        samples: Path to test directory with sample files.

    """
    from copy import deepcopy

    import yaml
    from aiida.plugins import DataFactory

    OTEPipelineData: "Type[OTEPipelineDataNode]" = DataFactory("execflow.otepipeline")
    declarative_pipeline_file: "Dict[str, Any]" = yaml.safe_load(
        (samples / "pipe.yml").read_bytes()
    )

    # Create Node from __init__ directly
    copy_file = deepcopy(declarative_pipeline_file)
    node = OTEPipelineData(
        version=copy_file["version"],
        strategies=copy_file["strategies"],
        pipelines=copy_file["pipelines"],
    )

    assert node.validate() is None
    assert node.version == int(declarative_pipeline_file["version"])
    assert node.pipelines == dict(declarative_pipeline_file["pipelines"])
    assert node.strategies == list(declarative_pipeline_file["strategies"])

    # Create Node through setting properties individually after initializing an "empty"
    # Node.
    copy_file = deepcopy(declarative_pipeline_file)
    node = OTEPipelineData()
    node.version = copy_file["version"]
    node.strategies = copy_file["strategies"]
    node.pipelines = copy_file["pipelines"]

    assert node.validate() is None
    assert node.version == int(declarative_pipeline_file["version"])
    assert node.pipelines == dict(declarative_pipeline_file["pipelines"])
    assert node.strategies == list(declarative_pipeline_file["strategies"])


@pytest.mark.parametrize(
    "input_keys",
    [
        ("strategies", "pipelines"),
        ("version", "strategies", "pipelines"),
    ],
    ids=["strat+pipe", "ver+strat+pipe"],
)
def test_validate_valid_cases(input_keys: "Tuple[str, ...]", samples: "Path") -> None:
    """Test various valid cases for validation.

    Parameters:
        input_keys: Input keys for OTEPipelineData from declarative pipeline file.
        samples: Path to test directory with sample files.

    """
    from copy import deepcopy

    import yaml
    from aiida.plugins import DataFactory

    OTEPipelineData: "Type[OTEPipelineDataNode]" = DataFactory("execflow.otepipeline")
    declarative_pipeline_file: "Dict[str, Any]" = yaml.safe_load(
        (samples / "pipe.yml").read_bytes()
    )

    node = OTEPipelineData(
        **{key: deepcopy(declarative_pipeline_file[key]) for key in input_keys}
    )
    assert node.validate() is None


@pytest.mark.parametrize(
    "input_keys",
    [
        ("strategies",),
        ("pipelines",),
        ("version",),
        ("version", "strategies"),
        ("version", "pipelines"),
    ],
    ids=["strat", "pipe", "ver", "ver+strat", "ver+pipe"],
)
def test_validate_invalid_cases(input_keys: "Tuple[str, ...]", samples: "Path") -> None:
    """Test various invalid cases for validation.

    Parameters:
        input_keys: Input keys for OTEPipelineData from declarative pipeline file.
        samples: Path to test directory with sample files.

    """
    from copy import deepcopy

    import yaml
    from aiida.common.exceptions import ValidationError
    from aiida.plugins import DataFactory

    OTEPipelineData: "Type[OTEPipelineDataNode]" = DataFactory("execflow.otepipeline")
    declarative_pipeline_file: "Dict[str, Any]" = yaml.safe_load(
        (samples / "pipe.yml").read_bytes()
    )

    node = OTEPipelineData(
        **{key: deepcopy(declarative_pipeline_file[key]) for key in input_keys}
    )

    exception_message_end = ""
    if "strategies" not in input_keys:
        exception_message_end += " strategies"
    if "strategies" not in input_keys and "pipelines" not in input_keys:
        exception_message_end += " and"
    if "pipelines" not in input_keys:
        exception_message_end += " pipelines"

    assert node.validate(strict=False) is None
    with pytest.raises(
        ValidationError, match=rf"^Cannot validate, missing{exception_message_end}$"
    ):
        node.validate(strict=True)



"""OTEAPI Download, Parse, and Resource strategy config AiiDA Data Node class."""
# pylint: disable=line-too-long,invalid-name
from typing import TYPE_CHECKING

from aiida.common.exceptions import ValidationError
from oteapi.models import ResourceConfig

from execflow.wrapper.data.genericconfig import GenericConfigData

if TYPE_CHECKING:  # pragma: no cover
    from typing import Any, Optional, Union


class ResourceConfigData(GenericConfigData):
    """Resource Strategy Data Configuration.

    Important:
        Either of the pairs of attributes `downloadUrl`/`mediaType` or
        `accessUrl`/`accessService` MUST be specified.

    Args:
        downloadUrl (Optional[str]): Definition: The URL of the downloadable file in a
            given format. E.g. CSV file or RDF file.

            Usage: `downloadURL` *SHOULD* be used for the URL at
            which this distribution is available directly, typically through a HTTPS
            GET request or SFTP.
        mediaType (Optional[str]): The media type of the distribution as defined by IANA
            [[IANA-MEDIA-TYPES](https://www.w3.org/TR/vocab-dcat-2/#bib-iana-media-types)].

            Usage: This property *SHOULD* be used when the media type of the
            distribution is defined in IANA
            [[IANA-MEDIA-TYPES](https://www.w3.org/TR/vocab-dcat-2/#bib-iana-media-types)].
        accessUrl (Optional[str]): A URL of the resource that gives access to a
            distribution of the dataset. E.g. landing page, feed, SPARQL endpoint.

            Usage:

            - `accessURL` *SHOULD* be used for the URL of a service or location that
              can provide access to this distribution, typically through a Web form,
              query or API call.
            - `downloadURL` is preferred for direct links to downloadable resources.

        accessService (Optional[str]): A data service that gives access to the
            distribution of the dataset.
        license (Optional[str]): A legal document under which the distribution is made
            available.
        accessRights (Optional[str]): A rights statement that concerns how the
            distribution is accessed.
        publisher (Optional[str]): The entity responsible for making the resource/item
            available.

    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        downloadUrl: "Optional[str]" = None,
        mediaType: "Optional[str]" = None,
        accessUrl: "Optional[str]" = None,
        accessService: "Optional[str]" = None,
        license: "Optional[str]" = None,  # pylint: disable=redefined-builtin
        accessRights: "Optional[str]" = None,
        publisher: "Optional[str]" = None,
        **kwargs: "Any",
    ) -> None:
        if downloadUrl is None:
            downloadUrl = ResourceConfig.schema()["properties"]["downloadUrl"][
                "default"
            ]

        if mediaType is None:
            mediaType = ResourceConfig.schema()["properties"]["mediaType"]["default"]

        if accessUrl is None:
            accessUrl = ResourceConfig.schema()["properties"]["accessUrl"]["default"]

        if accessService is None:
            accessService = ResourceConfig.schema()["properties"]["accessService"][
                "default"
            ]

        if license is None:
            license = ResourceConfig.schema()["properties"]["license"]["default"]

        if accessRights is None:
            accessRights = ResourceConfig.schema()["properties"]["accessRights"][
                "default"
            ]

        if publisher is None:
            publisher = ResourceConfig.schema()["properties"]["publisher"]["default"]

        super().__init__(**kwargs)

        attr_dict = {
            "downloadUrl": downloadUrl,
            "mediaType": mediaType,
            "accessUrl": accessUrl,
            "accessService": accessService,
            "license": license,
            "accessRights": accessRights,
            "publisher": publisher,
        }

        self.base.attributes.set_many(attr_dict)

    @property
    def downloadUrl(self) -> "Union[str, None]":
        """Definition: The URL of the downloadable file in a given format. E.g. CSV
        file or RDF file.

        Usage: `downloadURL` *SHOULD* be used for the URL at
        which this distribution is available directly, typically through a HTTPS
        GET request or SFTP.
        """
        return self.base.attributes.get("downloadUrl")

    @downloadUrl.setter
    def downloadUrl(self, value: "Union[str, None]") -> None:
        self.set_attribute("downloadUrl", value)

    @property
    def mediaType(self) -> "Union[str, None]":
        """The media type of the distribution as defined by IANA
        [[IANA-MEDIA-TYPES](https://www.w3.org/TR/vocab-dcat-2/#bib-iana-media-types)].

        Usage: This property *SHOULD* be used when the media
        type of the distribution is defined in IANA
        [[IANA-MEDIA-TYPES](https://www.w3.org/TR/vocab-dcat-2/#bib-iana-media-types)].
        """
        return self.base.attributes.get("mediaType")

    @mediaType.setter
    def mediaType(self, value: "Union[str, None]") -> None:
        self.set_attribute("mediaType", value)

    @property
    def accessUrl(self) -> "Union[str, None]":
        """A URL of the resource that gives access to a distribution of
        the dataset. E.g. landing page, feed, SPARQL endpoint.

        Usage:

        - `accessURL` *SHOULD* be used for the URL of a service or location that
           can provide access to this distribution, typically through a Web form,
           query or API call.
        - `downloadURL` is preferred for direct links to downloadable resources.

        """
        return self.base.attributes.get("accessUrl")

    @accessUrl.setter
    def accessUrl(self, value: "Union[str, None]") -> None:
        self.set_attribute("accessUrl", value)

    @property
    def accessService(self) -> "Union[str, None]":
        """A data service that gives access to the distribution of the dataset."""
        return self.base.attributes.get("accessService")

    @accessService.setter
    def accessService(self, value: "Union[str, None]") -> None:
        self.set_attribute("accessService", value)

    @property
    def license(self) -> "Union[str, None]":
        """A legal document under which the distribution is made available."""
        return self.base.attributes.get("license")

    @license.setter
    def license(self, value: "Union[str, None]") -> None:
        self.set_attribute("license", value)

    @property
    def accessRights(self) -> "Union[str, None]":
        """A rights statement that concerns how the distribution is accessed."""
        return self.base.attributes.get("accessRights")

    @accessRights.setter
    def accessRights(self, value: "Union[str, None]") -> None:
        self.set_attribute("accessRights", value)

    @property
    def publisher(self) -> "Union[str, None]":
        """The entity responsible for making the resource/item available."""
        return self.base.attributes.get("publisher")

    @publisher.setter
    def publisher(self, value: "Union[str, None]") -> None:
        self.set_attribute("publisher", value)

    def _validate(self) -> bool:
        """
        Check if either of the pairs of attributes `downloadUrl`/`mediaType` or
        `accessUrl`/`accessService` is specified.

        Raises:
            aiida.common.exceptions.ValidationError: If a valid pair of attributes
                cannot be determined.

        Returns:
            Whether or not the validation succeeded.

        """
        if (self.downloadUrl is None or self.mediaType is None) and (
            self.accessUrl is None or self.accessService is None
        ):
            raise ValidationError(
                "Either of the pairs of attributes downloadUrl/mediaType or "
                "accessUrl/accessService MUST be specified."
            )

        super()._validate()

        return True



"""Generic OTEAPI Config AiiDA Data Node class."""
from typing import TYPE_CHECKING

from aiida.common.exceptions import ModificationNotAllowed
from aiida.orm import Data

if TYPE_CHECKING:  # pragma: no cover
    from typing import Any


class GenericConfigData(Data):
    """Generic class for configuration objects.

    Args:
        configuration (dict): Model-specific configuration options,
            which can either be given as key/value-pairs or set as attributes.
        description (str): A description of the configuration model.

    """

    def __init__(
        self, configuration: "dict[str, Any]", description: str, **kwargs: "Any"
    ) -> None:
        super().__init__(**kwargs)

        attr_dict = {"configuration": configuration, "description": description}

        self.base.attributes.set_many(attr_dict)

    @property
    def configuration(self) -> "dict[str, Any]":
        """Model-specific configuration options, which can either be given as
        key/value-pairs or set as attributes."""
        return self.base.attributes.get("configuration")

    @configuration.setter
    def configuration(self, value: "dict[str, Any]") -> None:
        self.set_attribute("configuration", value)

    @property
    def description(self) -> str:
        """A description of the configuration model."""
        return self.base.attributes.get("description")

    @description.setter
    def description(self, value: str) -> None:
        self.set_attribute("description", value)

    def set_attribute(self, attribute_name: str, value: "Any") -> None:
        """Set an attribute, ensuring the Node is not yet stored.

        Args:
            attribute_name: The name of the attribute to set.
            value: The value of the attribute.

        """
        if self.is_stored:
            raise ModificationNotAllowed(
                f"The {self.__class__.__name__} object cannot be modified, "
                "it has already been stored."
            )

        self.base.attributes.set(attribute_name, value)



"""AiiDA Data Node classes."""
from .filterconfig import FilterConfigData
from .functionconfig import FunctionConfigData
from .mappingconfig import MappingConfigData
from .resourceconfig import ResourceConfigData
from .transformationconfig import TransformationConfigData, TransformationStatusData

__all__ = (
    "FilterConfigData",
    "FunctionConfigData",
    "MappingConfigData",
    "ResourceConfigData",
    "TransformationConfigData",
    "TransformationStatusData",
)



"""OTEAPI Mapping strategy config AiiDA Data Node class."""
# pylint: disable=invalid-name
from typing import TYPE_CHECKING

from oteapi.models import MappingConfig

from execflow.wrapper.data.genericconfig import GenericConfigData

if TYPE_CHECKING:  # pragma: no cover
    from typing import Any, Optional, Union

    RDFTriple = tuple[str, str, str]


class MappingConfigData(GenericConfigData):
    """Mapping Strategy Data Configuration.

    Args:
        mappingType (str): Type of registered mapping strategy.
        prefixes (Optional[dict[str, str]]): List of shortnames that expands to an
            IRI given as local value/IRI-expansion-pairs.
        triples (Optional[list[RDFTriple]]): List of RDF triples given as (subject,
            predicate, object).

    """

    def __init__(
        self,
        mappingType: str,
        prefixes: "Optional[dict[str, str]]" = None,
        triples: "Optional[list[RDFTriple]]" = None,
        **kwargs: "Any",
    ) -> None:
        if prefixes is None:
            prefixes = MappingConfig.schema()["properties"]["prefixes"]["default"]

        if triples is None:
            triples = MappingConfig.schema()["properties"]["triples"]["default"]

        super().__init__(**kwargs)

        attr_dict = {
            "mappingType": mappingType,
            "prefixes": prefixes,
            "triples": triples,
        }

        self.base.attributes.set_many(attr_dict)

    @property
    def mappingType(self) -> str:
        """Type of registered mapping strategy."""
        return self.base.attributes.get("mappingType")

    @mappingType.setter
    def mappingType(self, value: str) -> None:
        self.set_attribute("mappingType", value)

    @property
    def prefixes(self) -> "Union[dict[str, str], None]":
        """List of shortnames that expands to an IRI given as local
        value/IRI-expansion-pairs."""
        return self.base.attributes.get("prefixes")

    @prefixes.setter
    def prefixes(self, value: "Union[dict[str, str], None]") -> None:
        self.set_attribute("prefixes", value)

    @property
    def triples(self) -> "Union[RDFTriple, None]":
        """List of RDF triples given as (subject, predicate, object)."""
        return self.base.attributes.get("triples")

    @triples.setter
    def triples(self, value: "Union[RDFTriple, None]") -> None:
        self.set_attribute("triples", value)



"""OTEAPI Transformation strategy config AiiDA Data Node class."""
# pylint: disable=invalid-name
from typing import TYPE_CHECKING

from aiida.common.exceptions import ModificationNotAllowed
from aiida.orm import Data
from oteapi.models.transformationconfig import (
    TransformationConfig,
    TransformationStatus,
)

from execflow.wrapper.data.genericconfig import GenericConfigData

if TYPE_CHECKING:  # pragma: no cover
    from datetime import datetime
    from typing import Any, Optional, Union

    from oteapi.models.transformationconfig import ProcessPriority


class TransformationConfigData(GenericConfigData):
    """Transformation Strategy Data Configuration.

    Args:
        transformationType (str): Type of registered transformation strategy.
            E.g., `celery/remote`.
        name (Optional[str]): Human-readable name of the transformation strategy.
        due (Optional[datetime]): Optional field to indicate a due data/time for when a
            transformation should finish.
        priority (Optional[ProcessPriority]): Define the process priority of the
            transformation execution.
        secret (Optional[str]): Authorization secret given when running a
            transformation.

    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        transformationType: str,
        name: "Optional[str]" = None,
        due: "Optional[datetime]" = None,
        priority: "Optional[ProcessPriority]" = None,
        secret: "Optional[str]" = None,
        **kwargs: "Any",
    ) -> None:
        if name is None:
            name = TransformationConfig.schema()["properties"]["name"]["default"]

        if due is None:
            due = TransformationConfig.schema()["properties"]["due"]["default"]

        if priority is None:
            priority = TransformationConfig.schema()["properties"]["priority"][
                "default"
            ]

        if secret is None:
            secret = TransformationConfig.schema()["properties"]["secret"]["default"]

        super().__init__(**kwargs)

        attr_dict = {
            "transformationType": transformationType,
            "name": name,
            "due": due,
            "priority": priority,
            "secret": secret,
        }

        self.base.attributes.set_many(attr_dict)

    @property
    def transformationType(self) -> str:
        """Type of registered transformation strategy. E.g., `celery/remote`."""
        return self.base.attributes.get("transformationType")

    @transformationType.setter
    def transformationType(self, value: str) -> None:
        self.set_attribute("transformationType", value)

    @property
    def name(self) -> "Union[str, None]":
        """Human-readable name of the transformation strategy."""
        return self.base.attributes.get("name")

    @name.setter
    def name(self, value: "Union[str, None]") -> None:
        self.set_attribute("name", value)

    @property
    def due(self) -> "Union[datetime, None]":
        """Optional field to indicate a due data/time for when a transformation should
        finish."""
        return self.base.attributes.get("due")

    @due.setter
    def due(self, value: "Union[datetime, None]") -> None:
        self.set_attribute("due", value)

    @property
    def priority(self) -> "Union[ProcessPriority, None]":
        """Define the process priority of the transformation execution."""
        return self.base.attributes.get("priority")

    @priority.setter
    def priority(self, value: "Union[ProcessPriority, None]") -> None:
        self.set_attribute("priority", value)

    @property
    def secret(self) -> "Union[str, None]":
        """Authorization secret given when running a transformation."""
        return self.base.attributes.get("secret")

    @secret.setter
    def secret(self, value: "Union[str, None]") -> None:
        self.set_attribute("secret", value)


class TransformationStatusData(Data):
    """Return from transformation status.

    Args:
        id (str): ID for the given transformation process.
        status (Optional[str]): Status for the transformation process.
        messages (Optional[list[str]]): Messages related to the transformation process.
        created (Optional[datetime]): Time of creation for the transformation process.
            Given in UTC.
        startTime (Optional[datetime]): Time when the transformation process started.
            Given in UTC.
        finishTime (Optional[datetime]): Time when the tranformation process finished.
            Given in UTC.

    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        id: str,  # pylint: disable=redefined-builtin
        status: "Optional[str]" = None,
        messages: "Optional[list[str]]" = None,
        created: "Optional[datetime]" = None,
        startTime: "Optional[datetime]" = None,
        finishTime: "Optional[datetime]" = None,
        **kwargs: "Any",
    ) -> None:
        if status is None:
            status = TransformationStatus.schema()["properties"]["status"]["default"]

        if messages is None:
            messages = TransformationStatus.schema()["properties"]["messages"][
                "default"
            ]

        if created is None:
            created = TransformationStatus.schema()["properties"]["created"]["default"]

        if startTime is None:
            startTime = TransformationStatus.schema()["properties"]["startTime"][
                "default"
            ]

        if finishTime is None:
            finishTime = TransformationStatus.schema()["properties"]["finishTime"][
                "default"
            ]

        super().__init__(**kwargs)

        attr_dict = {
            "id": id,
            "status": status,
            "messages": messages,
            "created": created,
            "startTime": startTime,
            "finishTime": finishTime,
        }

        self.base.attributes.set_many(attr_dict)

    @property
    def id(self) -> str:
        """ID for the given transformation process."""
        return self.base.attributes.get("id")

    @id.setter
    def id(self, value: str) -> None:
        self.set_attribute("id", value)

    @property
    def status(self) -> "Union[str, None]":
        """Status for the transformation process."""
        return self.base.attributes.get("status")

    @status.setter
    def status(self, value: "Union[str, None]") -> None:
        self.set_attribute("status", value)

    @property
    def messages(self) -> "Union[list[str], None]":
        """Messages related to the transformation process."""
        return self.base.attributes.get("messages")

    @messages.setter
    def messages(self, value: "Union[list[str], None]") -> None:
        self.set_attribute("messages", value)

    @property
    def created(self) -> "Union[datetime, None]":
        """Time of creation for the transformation process. Given in UTC."""
        return self.base.attributes.get("created")

    @created.setter
    def created(self, value: "Union[datetime, None]") -> None:
        self.set_attribute("created", value)

    @property
    def startTime(self) -> "Union[datetime, None]":
        """Time when the transformation process started. Given in UTC."""
        return self.base.attributes.get("startTime")

    @startTime.setter
    def startTime(self, value: "Union[datetime, None]") -> None:
        self.set_attribute("startTime", value)

    @property
    def finishTime(self) -> "Union[datetime, None]":
        """Time when the tranformation process finished. Given in UTC."""
        return self.base.attributes.get("finishTime")

    @finishTime.setter
    def finishTime(self, value: "Union[datetime, None]") -> None:
        self.set_attribute("finishTime", value)

    def set_attribute(  # pylint: disable=duplicate-code
        self, attribute_name: str, value: "Any"
    ) -> None:
        """Set an attribute, ensuring the Node is not yet stored.

        Args:
            attribute_name: The name of the attribute to set.
            value: The value of the attribute.

        """
        if self.is_stored:
            raise ModificationNotAllowed(
                f"The {self.__class__.__name__} object cannot be modified, "
                "it has already been stored."
            )

        self.base.attributes.set(attribute_name, value)



"""Utility functions and classes for usage in ExecFlow."""
# pylint: disable=no-name-in-module,no-self-argument
from typing import TYPE_CHECKING, Dict, List, Union

from oteapi.models import (
    FilterConfig,
    FunctionConfig,
    MappingConfig,
    ResourceConfig,
    TransformationConfig,
)
from pydantic import BaseModel, Field, ValidationError, validator

if TYPE_CHECKING:  # pragma: no cover
    from typing import Any, Type


class DeclarativeStrategyBase(BaseModel):
    """A mix-in class for standard methods."""

    def get_name(self) -> str:
        """Get the name of the strategy."""
        name_field = self.__class__.__name__[len("Declarative") :].lower()
        if not hasattr(self, name_field):
            raise ValueError(
                f"Cannot find strategy name field {name_field!r}. Note, this function "
                "does not work from the DeclarativeStrategyBase mix-in class directly!"
            )
        return getattr(self, name_field)

    def get_config(self) -> "Dict[str, Any]":
        """Get the model as a ready-to-parse config for AiiDA calculations."""
        name_field = self.__class__.__name__[len("Declarative") :].lower()
        if not hasattr(self, name_field):
            raise ValueError(
                f"Cannot find strategy name field {name_field!r}. Note, this function "
                "does not work from the DeclarativeStrategyBase mix-in class directly!"
            )
        return self.dict(exclude={name_field})

    def __hash__(self) -> int:
        return hash(repr(self))


class DeclarativeDataResource(DeclarativeStrategyBase, ResourceConfig):
    """Data model for data resource."""

    dataresource: str = Field(
        ...,
        description="Name for the strategy within the declarative pipeline.",
    )


class DeclarativeFilter(DeclarativeStrategyBase, FilterConfig):
    """Data model for filter."""

    filter: str = Field(
        ...,
        description="Name for the strategy within the declarative pipeline.",
    )


class DeclarativeFunction(DeclarativeStrategyBase, FunctionConfig):
    """Data model for function."""

    function: str = Field(
        ...,
        description="Name for the strategy within the declarative pipeline.",
    )


class DeclarativeMapping(DeclarativeStrategyBase, MappingConfig):
    """Data model for mapping."""

    mapping: str = Field(
        ...,
        description="Name for the strategy within the declarative pipeline.",
    )


class DeclarativeTransformation(DeclarativeStrategyBase, TransformationConfig):
    """Data model for transformation."""

    transformation: str = Field(
        ...,
        description="Name for the strategy within the declarative pipeline.",
    )


if TYPE_CHECKING:  # pragma: no cover
    DeclarativeStrategy = Union[
        DeclarativeDataResource,
        DeclarativeFilter,
        DeclarativeFunction,
        DeclarativeMapping,
        DeclarativeTransformation,
    ]


class DeclarativePipeline(BaseModel):
    """Data model for a declarative pipeline."""

    version: int = Field(1, description="The declarative pipeline syntax version.")
    strategies: List[
        Union[
            DeclarativeDataResource,
            DeclarativeFilter,
            DeclarativeFunction,
            DeclarativeMapping,
            DeclarativeTransformation,
        ]
    ] = Field(
        ...,
        description=(
            "List of strategies to be used in the pipeline(s), including their "
            "configuration."
        ),
    )
    pipelines: Dict[str, str] = Field(
        ...,
        description="Dict of the pipeline(s) defined as part of the overall pipeline.",
    )

    @validator("strategies", pre=True, each_item=False)
    def type_cast_strategies(
        cls, value: "List[Dict[str, Any]]"
    ) -> "List[DeclarativeStrategy]":
        """Sort strategies into "correct" types."""
        type_mapping: "Dict[str, Type[DeclarativeStrategy]]" = {
            "dataresource": DeclarativeDataResource,
            "filter": DeclarativeFilter,
            "function": DeclarativeFunction,
            "mapping": DeclarativeMapping,
            "transformation": DeclarativeTransformation,
        }
        type_casted_strategies: "List[DeclarativeStrategy]" = []
        for strategy in value:
            for strategy_type, strategy_cls in type_mapping.items():
                if strategy_type in strategy:
                    try:
                        type_casted_strategies.append(strategy_cls(**strategy))
                    except ValidationError as exc:
                        exc_message = str(exc).replace("\n", "\n  ")
                        raise ValueError(
                            "Strategy cannot be validated as a "
                            f"{strategy_cls.__name__!r}."
                            f"\n  strategy: {strategy}\n  {exc_message}"
                        ) from exc
                    else:
                        break
            else:
                raise ValueError(
                    "Strategy does not define a valid strategy type as key:"
                    f"\n  strategy: {strategy}\n  Valid types: {type_mapping}"
                )
        return type_casted_strategies

    @validator("pipelines")
    def ensure_steps_exist(
        cls, value: "Dict[str, str]", values: "Dict[str, Any]"
    ) -> "Dict[str, str]":
        """Ensure the listed steps in the pipelines exist within the same data model."""
        strategy_names = [
            strategy.get_name() for strategy in values.get("strategies", [])
        ]
        pipeline_names = list(value)
        for name, pipeline in value.items():
            pipeline_parts = [part.strip() for part in pipeline.split("|")]
            if not all(
                part in strategy_names + pipeline_names for part in pipeline_parts
            ):
                raise ValueError(
                    f"Part(s) in {name!r} pipeline not found in the list of strategies"
                    " or pipelines within the declarative pipeline."
                )
        return value

    def _strategy_names(self) -> "List[str]":
        """Utility function to return list of strategy names.

        Returns:
            List of strategy names as given in the declarative pipeline.

        """
        return [strategy.get_name() for strategy in self.strategies]

    def get_strategy(self, strategy_name: str) -> "DeclarativeStrategy":
        """Get a strategy based on its given name.

        Parameters:
            strategy_name: The name given the strategy in the declarative pipeline.

        Returns:
            The declarative strategy model object with the given name.

        """
        for strategy in self.strategies:
            if strategy_name == strategy.get_name():
                return strategy
        raise ValueError(
            f"Strategy {strategy_name!r} does not exist among {self._strategy_names()}"
        )

    def parse_pipeline(self, pipeline: str) -> "List[DeclarativeStrategy]":
        """Resolve a pipeline into its strategy parts.

        Parameters:
            pipeline: Name of the pipeline.

        Returns:
            (Ordered) list of strategies in the pipeline.

        """
        if pipeline not in self.pipelines:
            raise ValueError(
                f"Pipeline {pipeline!r} does not exist among {list(self.pipelines)}"
            )

        parsed_pipeline = []

        pipeline_parts = [part.strip() for part in self.pipelines[pipeline].split("|")]
        for part in pipeline_parts:
            if part in self._strategy_names():
                parsed_pipeline.append(self.get_strategy(part))
            elif part in self.pipelines:
                parsed_pipeline.extend(self.parse_pipeline(part))
            else:
                raise ValueError(
                    f"Could unexpectedly not find part {part!r} in the list of known "
                    "strategy and pipeline names!"
                )
        return parsed_pipeline

    def __hash__(self) -> int:
        return hash(repr(self))



"""AiiDA WorkChain for the OTE Pipeline."""
from pathlib import Path
from typing import TYPE_CHECKING

from aiida import orm
from aiida.engine import run_get_node, while_
from aiida.engine.processes.workchains.workchain import WorkChain

from execflow.wrapper import calculations
from execflow.wrapper.data.declarative_pipeline import OTEPipelineData

if TYPE_CHECKING:  # pragma: no cover
    from aiida.engine.processes.workchains.workchain import WorkChainSpec


strategy_calculation_mapping = {
    "DeclarativeDataResource": {
        "init": calculations.init_dataresource,
        "get": calculations.get_dataresource,
    },
    "DeclarativeFilter": {
        "init": calculations.init_filter,
        "get": calculations.get_filter,
    },
    "DeclarativeFunction": {
        "init": calculations.init_function,
        "get": calculations.get_function,
    },
    "DeclarativeMapping": {
        "init": calculations.init_mapping,
        "get": calculations.get_mapping,
    },
    "DeclarativeTransformation": {
        "init": calculations.init_transformation,
        "get": calculations.get_transformation,
    },
}


class OTEPipeline(WorkChain):
    """Run an OTE Pipeline.

    Inputs:
        - **pipeline**
          (:py:class:`~execflow.wrapper.data.declarative_pipeline.OTEPipelineData`,
          :py:class:`aiida.orm.Dict`, :py:class:`aiida.orm.SinglefileData`,
          :py:class:`aiida.orm.Str`) -- The declarative pipeline as an AiiDA-valid
          type. Either as a path to a YAML file or the explicit content of the YAML
          file.

    Outputs:
        - **session** (:py:class:`aiida.orm.Dict`) -- The OTE session object after
          running the pipeline.

    Outline:
        - :py:meth:`~execflow.wrapper.workflows.pipeline.OTEPipeline.setup`
        - while
          :py:meth:`~execflow.wrapper.workflows.pipeline.OTEPipeline.not_finished`:

          - :py:meth:`~execflow.wrapper.workflows.pipeline.OTEPipeline.submit_next`
          - :py:meth:`~execflow.wrapper.workflows.pipeline.OTEPipeline.process_current`

        - :py:meth:`~execflow.wrapper.workflows.pipeline.OTEPipeline.finalize`

    Exit Codes:
        - **2** (*ERROR_SUBPROCESS*) -- A subprocess has failed.

    """

    @classmethod
    def define(cls, spec: "WorkChainSpec") -> None:
        super().define(spec)

        spec.input(
            "pipeline",
            valid_type=(OTEPipelineData, orm.Dict, orm.SinglefileData, orm.Str),
        )
        spec.exit_code(2, "ERROR_SUBPROCESS", message="A subprocess has failed.")

        spec.outline(
            cls.setup,
            while_(cls.not_finished)(cls.submit_next, cls.process_current),
            cls.finalize,
        )
        spec.output("session", valid_type=orm.Dict)

    def setup(self) -> None:
        """Setup WorkChain

        Steps:

        - Initialize context.
        - Parse declarative pipeline.
        - Create a list of strategies to run, explicitly adding `init` and `get`
          CalcFunctions.

        """
        self.ctx.current_id = 0

        if isinstance(self.inputs["pipeline"], (OTEPipelineData, orm.Dict)):
            pipeline = OTEPipelineData(value=self.inputs["pipeline"])
        elif isinstance(self.inputs["pipeline"], orm.SinglefileData):
            pipeline = OTEPipelineData(single_file=self.inputs["pipeline"])
        elif isinstance(self.inputs["pipeline"], orm.Str):
            if Path(self.inputs["pipeline"].value).resolve().exists():
                pipeline = OTEPipelineData(filepath=self.inputs["pipeline"])
            else:
                pipeline = OTEPipelineData(value=self.inputs["pipeline"])

        # Outline pipeline
        # For the moment we can only handle a single defined pipeline
        if not pipeline.pipelines or len(pipeline.pipelines) > 1:
            raise NotImplementedError(
                "Currently there is not way of determining which pipeline to run. "
                "Please only supply a single pipeline in the 'pipelines' object of "
                "the declarative pipeline."
            )
        run_pipeline_name = list(pipeline.pipelines)[0]
        strategies = []
        # Initialization
        for strategy in pipeline.get_strategies(run_pipeline_name, reverse=True):
            strategies.append(
                (
                    strategy_calculation_mapping[strategy.__class__.__name__]["init"],
                    strategy.get_config(),
                )
            )
        # Getting
        for strategy in pipeline.get_strategies(run_pipeline_name):
            strategies.append(
                (
                    strategy_calculation_mapping[strategy.__class__.__name__]["get"],
                    strategy.get_config(),
                )
            )

        self.ctx.strategies = strategies

        self.ctx.ote_session = orm.Dict()

    def not_finished(self) -> bool:
        """Determine whether or not the WorkChain is finished.

        Returns:
            Whether or not the WorkChain is finished based on comparing the current
            strategy index in the list of strategies against the total number of
            strategies.

        """
        return self.ctx.current_id < len(self.ctx.strategies)

    def submit_next(self) -> None:
        """Prepare the current step for submission.

        Run the next strategy's CalcFunction and return its ProcessNode to the context.

        """
        current_id = self.ctx.current_id
        calc_function, strategy_config = self.ctx.strategies[current_id]
        session = self.ctx.ote_session

        self.to_context(
            current=run_get_node(
                calc_function, **{"config": strategy_config, "session": session}
            )[1]
        )

    def process_current(self) -> None:
        """Process the current step's Node.

        Report if the process did not finish OK.
        Retrieve the return session update object, update the session and store it back
        to the context for the next strategy to use.

        """
        if not self.ctx.current.is_finished_ok:
            self.report(
                f"A subprocess failed with exit status {self.ctx.current.exit_status}:"
                f" {self.ctx.current.exit_message}"
            )

        session = self.ctx.ote_session.get_dict()
        session.update(
            self.ctx.current.base.links.get_outgoing()
            .get_node_by_label("result")
            .get_dict()
        )
        self.ctx.ote_session = orm.Dict(session)

        self.ctx.current_id += 1

    def finalize(self) -> None:
        """Finalize the WorkChain.

        Set the 'session' output.
        """
        self.out("session", self.ctx.ote_session)



"""AiiDA Calculations for the OTE Data Resource Strategy."""
from aiida.engine import calcfunction
from aiida.orm import Dict
from oteapi.models import ResourceConfig
from oteapi.plugins import create_strategy, load_strategies


@calcfunction
def init_dataresource(config: Dict, session: Dict) -> Dict:
    """Initialize an OTE Data Resource strategy."""
    resource_config = ResourceConfig(**config.get_dict())

    load_strategies()

    if resource_config.downloadUrl and resource_config.mediaType:
        # Download strategy
        session_update = create_strategy("download", resource_config).initialize(
            session.get_dict()
        )

        # Parse strategy
        parse_session = session.get_dict()
        parse_session.update(session_update)
        session_update = create_strategy("parse", resource_config).initialize(
            session.get_dict()
        )
    elif resource_config.accessUrl and resource_config.accessService:
        # Resource strategy
        session_update = create_strategy("resource", resource_config).initialize(
            session.get_dict()
        )
    else:
        raise ValueError(
            "Either of the pairs downloadUrl/mediaType and accessUrl/accessService "
            "must be defined in the config."
        )
    return Dict(session_update)


@calcfunction
def get_dataresource(config: Dict, session: Dict) -> Dict:
    """Get/Execute an OTE Data Resource strategy."""
    resource_config = ResourceConfig(**config.get_dict())

    load_strategies()

    if resource_config.downloadUrl and resource_config.mediaType:
        # Download strategy
        session_update = create_strategy("download", resource_config).get(
            session.get_dict()
        )

        # Parse strategy
        parse_session = session.get_dict()
        parse_session.update(session_update)
        session_update = create_strategy("parse", resource_config).get(parse_session)
    elif resource_config.accessUrl and resource_config.accessService:
        # Resource strategy
        session_update = create_strategy("resource", resource_config).get(
            session.get_dict()
        )
    else:
        raise ValueError(
            "Either of the pairs downloadUrl/mediaType and accessUrl/accessService "
            "must be defined in the config."
        )
    return Dict(session_update)



"""A Data Node representing a declarative OTE pipeline."""
from copy import deepcopy
from pathlib import Path
from typing import TYPE_CHECKING

import yaml
from aiida.common.exceptions import NotExistent, ValidationError
from aiida.orm import Dict as DictNode
from aiida.orm import SinglefileData
from aiida.orm import Str as StrNode

from execflow.utils import DeclarativePipeline

if TYPE_CHECKING:  # pragma: no cover
    from typing import Any, Dict, Generator, Iterable, List, Optional, Tuple, Union

    from execflow.utils import DeclarativeStrategy


class OTEPipelineData(DictNode):
    """An OTE pipeline.

    The data structure is based on the declarative pipeline syntax.

    Strategy for initializing:

    - Either `value` or either of `filepath` and `single_file` are defined.
    - All explicit top keywords either modify now existing top keywords or create them.

    """

    _pydantic_model_class = DeclarativePipeline
    _pydantic_model_hash = None
    _pydantic_model = None

    def __init__(  # pylint: disable=too-many-branches,too-many-statements
        self,
        # From dict
        value: "Optional[Union[Dict[str, Any], OTEPipelineData, DictNode, DeclarativePipeline, str, bytes, StrNode]]" = None,  # pylint: disable=line-too-long
        # From file
        filepath: "Optional[Union[Path, str, StrNode]]" = None,
        single_file: "Optional[SinglefileData]" = None,
        # Explicitly - top keywords
        version: "Optional[Union[int, str]]" = None,
        strategies: "Optional[List[Dict[str, Any]]]" = None,
        pipelines: "Optional[Dict[str, str]]" = None,
        # AiiDA-specific extra keyword-arguments
        **kwargs: "Any",
    ) -> None:
        # Update
        dictionary = None

        if value is not None:
            dictionary = deepcopy(value)

        if dictionary is None:
            if filepath and single_file:
                raise ValueError(
                    "Specify either 'filepath' or 'single_file', not both."
                )
            if filepath:
                filepath = (
                    Path(filepath.value).resolve()
                    if isinstance(filepath, StrNode)
                    else Path(filepath).resolve()
                )
                if not filepath.exists():
                    raise FileNotFoundError(f"Cannot find file at filepath={filepath}")
                try:
                    dictionary = yaml.safe_load(filepath.read_bytes())
                except yaml.error.YAMLError as exc:
                    raise ValueError(
                        "Could not parse declarative pipeline YAML file from "
                        f"filepath={filepath}."
                    ) from exc
            elif single_file:
                try:
                    dictionary = yaml.safe_load(single_file.get_content())
                except yaml.error.YAMLError as exc:
                    raise ValueError(
                        "Could not parse declarative pipeline YAML file from "
                        f"single_file={single_file}."
                    ) from exc

        if dictionary and isinstance(dictionary, (str, bytes, StrNode)):
            try:
                dictionary = (
                    yaml.safe_load(dictionary.value)
                    if isinstance(dictionary, StrNode)
                    else yaml.safe_load(dictionary)
                )
            except yaml.error.YAMLError as exc:
                raise ValueError(
                    "Could not parse declarative pipeline YAML file from 'value'."
                ) from exc

        if dictionary and isinstance(dictionary, self._pydantic_model_class):
            dictionary = dictionary.dict()

        dictionary = dictionary or {}

        if "version" in dictionary:
            try:
                dictionary["version"] = int(dictionary["version"])
            except ValueError as exc:
                raise ValueError(
                    f"{dictionary['version']!r} is not a valid value for 'version'."
                ) from exc

        if version is not None:
            try:
                version = int(version)
            except ValueError as exc:
                raise ValueError(
                    f"{version!r} is not a valid value for 'version'."
                ) from exc
            dictionary["version"] = version

        if strategies is not None:
            if "strategies" in dictionary:
                original_strategies = dictionary["strategies"]
                for strategy in list(strategies):
                    if strategy in original_strategies:
                        strategies.remove(strategy)
                dictionary["strategies"] = dictionary["strategies"].extend(strategies)
            else:
                dictionary["strategies"] = strategies

        if pipelines is not None:
            if "pipelines" in dictionary:
                updated_pipelines: "Dict[str, str]" = dictionary["pipelines"]
                updated_pipelines.update(pipelines)
                dictionary["pipelines"] = updated_pipelines
            else:
                dictionary["pipelines"] = pipelines

        super().__init__(value=dictionary, **kwargs)
        self.validate()

    def validate(self, strict: bool = False) -> None:
        """Validate the internal data model via the pydantic data model.

        Parameters:
            strict: Whether or not to raise if `strategies` or `pipelines` are None.

        """
        if self.strategies is not None and self.pipelines is not None:
            self._pydantic_model = self._pydantic_model_class(**self.get_dict())
            self._pydantic_model_hash = hash(self._pydantic_model)
        elif strict:
            raise ValidationError(
                "Cannot validate, missing"
                f"{' strategies' if self.strategies is None else ''}"
                f"{' and' if self.strategies is None and self.pipelines is None else ''}"  # pylint: disable=line-too-long
                f"{' pipelines' if self.pipelines is None else ''}"
            )

    @property
    def version(self) -> "Optional[int]":
        """Return the declarative pipeline syntax version.

        Returns:
            The declarative pipeline syntax version or `None` if `version` is not set.

        """
        return self.base.attributes.get("version", None)

    @version.setter
    def version(self, value: int) -> None:
        """Set the version.

        Parameters:
            value: The version to store.

        """
        try:
            value = int(value)
        except (TypeError, ValueError) as exc:
            raise TypeError("'value' must be an integer.") from exc
        self["version"] = value

    @property
    def strategies(self) -> "Optional[List[Dict[str, Any]]]":
        """Return a stand-alone list of strategies.

        By "stand-alone" it is meant that mutating the returned list, will not result
        in changes in the strategies stored within the Node.

        Example:
            To extend the strategies in the OTEPipelineData Node `node` with a list
            of strategies `my_extra_strategies`, you can do:

            .. code:: python

               node["strategies"] = node["strategies"].extend(my_extra_strategies)

            or

            .. code:: python

                node.dict.strategies = node.dict.strategies.extend(my_extra_strategies)

            or a mix of these approaches.

        Returns:
            The list of strategies or `None` if no strategies have been set.

        """
        return self.base.attributes.get("strategies", None)

    @strategies.setter
    def strategies(self, value: "Iterable[Dict[str, Any]]") -> None:
        """Set the list of strategies.

        Parameters:
            value: The list of strategies to store.

        """
        try:
            value = list(value)
        except (TypeError, ValueError) as exc:
            raise TypeError("'value' must be an iterable of dictionaries.") from exc
        self["strategies"] = value

    @property
    def pipelines(self) -> "Optional[Dict[str, Any]]":
        """Return a stand-alone dictionary of pipelines.

        By "stand-alone" it is meant that mutating the returned dictionary, will not
        result in changes in the pipelines stored within the Node.

        Example:
            To update the pipelines in the OTEPipelineData Node `node` with a
            dictionary of pipelines `my_updated_pipelines`, you can do:

            .. code:: python

                updated_pipelines = node["pipelines"]
                updated_pipelines.update(my_updated_pipelines)
                node["pipelines"] = updated_pipelines

            or

            .. code:: python

                updated_pipelines = node.dict.pipelines
                updated_pipelines.update(my_updated_pipelines)
                node.dict.pipelines = updated_pipelines

            or a mix of these approaches.

            Note, from Python 3.9 the above example can be simplified to:

            .. code:: python

                node.dict.pipelines = node["pipelines"] | my_updated_pipelines

        Returns:
            The pipelines or `None` if no pipelines have been set.

        """
        return self.base.attributes.get("pipelines", None)

    @pipelines.setter
    def pipelines(
        self, value: "Union[Dict[str, str], Iterable[Tuple[str, str]]]"
    ) -> None:
        """Set the dictionary of pipelines.

        Parameters:
            value: The dictionary of pipelines to store.

        """
        try:
            value = dict(value)
        except (TypeError, ValueError) as exc:
            raise TypeError("'value' must be a dictionary.") from exc
        self["pipelines"] = value

    @property
    def pydantic_model(self) -> "DeclarativePipeline":
        """Instantiate and return a pydantic model of the declarative pipeline.

        Returns:
            A pydantic model representing the declarative pipeline.

        """
        if self.strategies is None or self.pipelines is None:
            raise NotExistent(
                "Cannot instantiate a pydantic model, missing"
                f"{' strategies' if self.strategies is None else ''}"
                f"{' and' if self.strategies is None and self.pipelines is None else ''}"  # pylint: disable=line-too-long
                f"{' pipelines' if self.pipelines is None else ''}"
            )

        if self._pydantic_model_hash is None or self._pydantic_model_hash != hash(
            self._pydantic_model
        ):
            self._pydantic_model = self._pydantic_model_class(**self.get_dict())
            self._pydantic_model_hash = hash(self._pydantic_model)

        return self._pydantic_model

    def get_strategies(
        self, pipeline: str, reverse: bool = False
    ) -> "Generator[DeclarativeStrategy, None, None]":
        """Yield pipeline strategies as pydantic models.

        Parameters:
            pipeline: The pipeline name to iterate over / resolve.
            reverse: Whether or not to iterate over / resolve the pipeline in reverse.

        Yields:
            A declarative strategy model object.

        """
        self.validate(strict=True)
        if self.pipelines and pipeline not in self.pipelines:
            raise NotExistent(
                f"Pipeline {pipeline!r} does not exist among {list(self.pipelines)}"
            )

        if reverse:
            for strategy in reversed(self.pydantic_model.parse_pipeline(pipeline)):
                yield strategy
        else:
            for strategy in self.pydantic_model.parse_pipeline(pipeline):
                yield strategy



"""Repository management tasks powered by `invoke`.
More information on `invoke` can be found at http://www.pyinvoke.org/.
"""
# pylint: disable=import-outside-toplevel
import re
import sys
from pathlib import Path
from typing import TYPE_CHECKING

from invoke import task

if TYPE_CHECKING:  # pragma: no cover
    from typing import Optional, Tuple

    from invoke import Context, Result


TOP_DIR = Path(__file__).parent.resolve()


def update_file(
    filename: Path, sub_line: "Tuple[str, str]", strip: "Optional[str]" = None
) -> None:
    """Utility function for tasks to read, update, and write files"""
    lines = [
        re.sub(sub_line[0], sub_line[1], line.rstrip(strip))
        for line in filename.read_text(encoding="utf8").splitlines()
    ]
    filename.write_text("\n".join(lines) + "\n", encoding="utf8")


@task(help={"ver": "ExecFlow SDK version to set"})
def setver(_, ver=""):
    """Sets the ExecFlow SDK version."""
    match = re.fullmatch(
        (
            r"v?(?P<version>[0-9]+(\.[0-9]+){2}"  # Major.Minor.Patch
            r"(-[0-9A-Za-z-]+(\.[0-9A-Za-z-]+)*)?"  # pre-release
            r"(\+[0-9A-Za-z-]+(\.[0-9A-Za-z-]+)*)?)"  # build metadata
        ),
        ver,
    )
    if not match:
        sys.exit(
            "Error: Please specify version as "
            "'Major.Minor.Patch(-Pre-Release+Build Metadata)' or "
            "'vMajor.Minor.Patch(-Pre-Release+Build Metadata)'"
        )
    ver = match.group("version")

    update_file(
        TOP_DIR / "execflow" / "__init__.py",
        (r'__version__ = (\'|").*(\'|")', f'__version__ = "{ver}"'),
    )

    print(f"Bumped version to {ver}.")


@task(
    help={
        "fail-fast": (
            "Fail immediately if an error occurs. Otherwise, print and ignore all "
            "non-critical errors."
        ),
    }
)
def update_deps(  # pylint: disable=too-many-branches,too-many-locals,too-many-statements
    context, fail_fast=False
):
    """Update dependencies in `pyproject.toml`."""
    import tomlkit

    if TYPE_CHECKING:  # pragma: no cover
        context: "Context" = context

    pyproject_path = TOP_DIR / "pyproject.toml"
    pyproject = tomlkit.loads(pyproject_path.read_bytes())

    py_version = re.match(
        r"^.*(?P<version>3\.[0-9]+)$",
        pyproject.get("project", {}).get("requires-python", ""),
    ).group("version")

    already_handled_packages = []
    updated_packages = {}
    dependencies = pyproject.get("project", {}).get("dependencies", [])
    for optional_deps in (
        pyproject.get("project", {}).get("optional-dependencies", {}).values()
    ):
        dependencies.extend(optional_deps)
    for line in dependencies:
        match = re.match(
            r"^(?P<full_dependency>(?P<package>[a-zA-Z0-9-_]+)\S*) "
            r"(?P<operator>>|<|<=|>=|==|!=|~=)"
            r"(?P<version>[0-9]+(?:\.[0-9]+){0,2})"
            r"(?P<extra_op_ver>(?:,(?:>|<|<=|>=|==|!=|~=)"
            r"[0-9]+(?:\.[0-9]+){0,2})*)$",
            line,
        )
        if match is None:
            msg = f"Could not parse package, operator, and version for line:\n  {line}"
            if fail_fast:
                sys.exit(msg)
            print(msg)
        full_dependency_name = match.group("full_dependency")
        package = match.group("package")
        operator = match.group("operator")
        version = match.group("version")
        extra_op_ver = match.group("extra_op_ver")

        # Skip package if already handled
        if package in already_handled_packages:
            continue

        out: "Result" = context.run(
            f"pip index versions --python-version {py_version} {package}",
            hide=True,
        )
        package_latest_version_line = out.stdout.split(sep="\n", maxsplit=1)[0]
        match = re.match(
            r"(?P<package>[a-zA-Z0-9-_]+) \((?P<version>[0-9]+(?:\.[0-9]+){0,2})\)",
            package_latest_version_line,
        )
        if match is None:
            msg = (
                "Could not parse package and version from 'pip index versions' output "
                f"for line:\n  {package_latest_version_line}"
            )
            if fail_fast:
                sys.exit(msg)
            print(msg)

        # Sanity check
        if package != match.group("package"):
            msg = (
                f"Package name parsed from pyproject.toml ({package!r}) does not match"
                " the name returned from 'pip index versions': "
                f"{match.group('package')!r}"
            )
            if fail_fast:
                sys.exit(msg)
            print(msg)

        latest_version = match.group("version").split(".")
        for index, version_part in enumerate(version.split(".")):
            if version_part != latest_version[index]:
                break
        else:
            already_handled_packages.append(package)
            continue

        # Update pyproject.toml
        updated_version = ".".join(latest_version[: len(version.split("."))])
        escaped_full_dependency_name = full_dependency_name.replace(
            "[", "\["  # pylint: disable=anomalous-backslash-in-string
        ).replace(
            "]", "\]"  # pylint: disable=anomalous-backslash-in-string
        )
        update_file(
            pyproject_path,
            (
                rf'"{escaped_full_dependency_name} {operator}.*"',
                f'"{full_dependency_name} {operator}{updated_version}'
                f'{extra_op_ver if extra_op_ver else ""}"',
            ),
        )
        already_handled_packages.append(package)
        updated_packages[full_dependency_name] = f"{operator}{updated_version}"

    if updated_packages:
        print(
            "Successfully updated the following dependencies:\n"
            + "\n".join(
                f"  {package} ({version}{extra_op_ver if extra_op_ver else ''})"
                for package, version in updated_packages.items()
            )
            + "\n"
        )
    else:
        print("No dependency updates available.")



"""AiiDA Calculations representing OTE strategies."""
from .dataresource import get_dataresource, init_dataresource
from .filter import get_filter, init_filter
from .function import get_function, init_function
from .mapping import get_mapping, init_mapping
from .transformation import get_transformation, init_transformation

__all__ = (
    "init_dataresource",
    "get_dataresource",
    "init_filter",
    "get_filter",
    "init_function",
    "get_function",
    "init_mapping",
    "get_mapping",
    "init_transformation",
    "get_transformation",
)



"""AiiDA Calculations for the OTE Mapping Strategy."""
from typing import TYPE_CHECKING

from aiida.engine import calcfunction
from aiida.orm import Dict
from oteapi.models import MappingConfig
from oteapi.plugins import create_strategy, load_strategies

if TYPE_CHECKING:  # pragma: no cover
    from oteapi.interfaces import IMappingStrategy


@calcfunction
def init_mapping(config: Dict, session: Dict) -> Dict:
    """Initialize an OTE Mapping strategy."""
    load_strategies()

    mapping_config = MappingConfig(**config.get_dict())
    strategy: "IMappingStrategy" = create_strategy("mapping", mapping_config)
    return Dict(strategy.initialize(session.get_dict()))


@calcfunction
def get_mapping(config: Dict, session: Dict) -> Dict:
    """Get/Execute an OTE Mapping strategy."""
    load_strategies()

    mapping_config = MappingConfig(**config.get_dict())
    strategy: "IMappingStrategy" = create_strategy("mapping", mapping_config)
    return Dict(strategy.get(session.get_dict()))



"""OTEAPI Function strategy config AiiDA Data Node class."""
# pylint: disable=invalid-name
from typing import TYPE_CHECKING

from execflow.wrapper.data.genericconfig import GenericConfigData

if TYPE_CHECKING:  # pragma: no cover
    from typing import Any


class FunctionConfigData(GenericConfigData):
    """Function Strategy Data Configuration.

    Args:
        functionType (str): Type of registered function strategy.

    """

    def __init__(self, functionType: str, **kwargs: "Any") -> None:
        super().__init__(**kwargs)

        self.base.attributes.set("functionType", functionType)

    @property
    def functionType(self) -> str:
        """Type of registered function strategy."""
        return self.base.attributes.get("functionType")

    @functionType.setter
    def functionType(self, value: str) -> None:
        self.set_attribute("functionType", value)
